@@ -1,6 +1,6 @@
-/* Copyright (c) 2015-2020 The Khronos Group Inc.
- * Copyright (c) 2015-2020 Valve Corporation
- * Copyright (c) 2015-2020 LunarG, Inc.
+/* Copyright (c) 2015-2021 The Khronos Group Inc.
+ * Copyright (c) 2015-2021 Valve Corporation
+ * Copyright (c) 2015-2021 LunarG, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -84,15 +84,14 @@ typedef enum DebugCallbackStatusBits {
 typedef VkFlags DebugCallbackStatusFlags;
 
 struct LogObjectList {
-    std::vector<VulkanTypedHandle> object_list;
+    small_vector<VulkanTypedHandle, 4, uint32_t> object_list;
 
     template <typename HANDLE_T>
     void add(HANDLE_T object) {
-        VulkanTypedHandle v_t_handle(object, ConvertCoreObjectToVulkanObject(VkHandleInfo<HANDLE_T>::kVkObjectType));
-        object_list.push_back(v_t_handle);
+        object_list.emplace_back(object, ConvertCoreObjectToVulkanObject(VkHandleInfo<HANDLE_T>::kVkObjectType));
     }
 
-    void add(VulkanTypedHandle typed_handle) { object_list.push_back(typed_handle); }
+    void add(VulkanTypedHandle typed_handle) { object_list.emplace_back(typed_handle); }
 
     template <typename HANDLE_T>
     LogObjectList(HANDLE_T object) {
@@ -118,9 +117,9 @@ typedef struct VkLayerDbgFunctionState {
 
     void *pUserData;
 
-    bool IsUtils() { return ((callback_status & DEBUG_CALLBACK_UTILS) != 0); }
-    bool IsDefault() { return ((callback_status & DEBUG_CALLBACK_DEFAULT) != 0); }
-    bool IsInstance() { return ((callback_status & DEBUG_CALLBACK_INSTANCE) != 0); }
+    bool IsUtils() const { return ((callback_status & DEBUG_CALLBACK_UTILS) != 0); }
+    bool IsDefault() const { return ((callback_status & DEBUG_CALLBACK_DEFAULT) != 0); }
+    bool IsInstance() const { return ((callback_status & DEBUG_CALLBACK_INSTANCE) != 0); }
 } VkLayerDbgFunctionState;
 
 // TODO: Could be autogenerated for the specific handles for extra type safety...
@@ -140,7 +139,7 @@ struct LoggingLabel {
     bool Empty() const { return name.empty(); }
 
     VkDebugUtilsLabelEXT Export() const {
-        auto out = lvl_init_struct<VkDebugUtilsLabelEXT>();
+        auto out = LvlInitStruct<VkDebugUtilsLabelEXT>();
         out.pLabelName = name.c_str();
         std::copy(color.cbegin(), color.cend(), out.color);
         return out;
@@ -197,9 +196,12 @@ typedef struct _debug_report_data {
     std::unordered_map<uint64_t, std::string> debugUtilsObjectNameMap;
     std::unordered_map<VkQueue, std::unique_ptr<LoggingLabelState>> debugUtilsQueueLabels;
     std::unordered_map<VkCommandBuffer, std::unique_ptr<LoggingLabelState>> debugUtilsCmdBufLabels;
+    std::vector<uint32_t> filter_message_ids{};
     // This mutex is defined as mutable since the normal usage for a debug report object is as 'const'. The mutable keyword allows
     // the layers to continue this pattern, but also allows them to use/change this specific member for synchronization purposes.
     mutable std::mutex debug_output_mutex;
+    int32_t duplicate_message_limit = 0;
+    mutable std::unordered_map<uint32_t, int32_t> duplicate_message_count_map{};
     const void *instance_pnext_chain{};
 
     void DebugReportSetUtilsObjectName(const VkDebugUtilsObjectNameInfoEXT *pNameInfo) {
@@ -298,7 +300,7 @@ static inline bool debug_log_msg(const debug_report_data *debug_data, VkFlags ms
 
 static void SetDebugUtilsSeverityFlags(std::vector<VkLayerDbgFunctionState> &callbacks, debug_report_data *debug_data) {
     // For all callback in list, return their complete set of severities and modes
-    for (auto item : callbacks) {
+    for (const auto &item : callbacks) {
         if (item.IsUtils()) {
             debug_data->active_severities |= item.debug_utils_msg_flags;
             debug_data->active_types |= item.debug_utils_msg_type;
@@ -333,6 +335,22 @@ static inline void RemoveAllMessageCallbacks(debug_report_data *debug_data, std:
     callbacks.clear();
 }
 
+// Returns TRUE if the number of times this message has been logged is over the set limit
+static inline bool UpdateLogMsgCounts(const debug_report_data *debug_data, int32_t vuid_hash) {
+    auto vuid_count_it = debug_data->duplicate_message_count_map.find(vuid_hash);
+    if (vuid_count_it == debug_data->duplicate_message_count_map.end()) {
+        debug_data->duplicate_message_count_map.insert({vuid_hash, 1});
+        return false;
+    } else {
+        if (vuid_count_it->second >= debug_data->duplicate_message_limit) {
+            return true;
+        } else {
+            vuid_count_it->second++;
+            return false;
+        }
+    }
+}
+
 static inline bool debug_log_msg(const debug_report_data *debug_data, VkFlags msg_flags, const LogObjectList &objects,
                                  const char *layer_prefix, const char *message, const char *text_vuid) {
     bool bail = false;
@@ -383,10 +401,14 @@ static inline bool debug_log_msg(const debug_report_data *debug_data, VkFlags ms
         }
     }
 
-    size_t location = 0;
+    int32_t location = 0;
     if (text_vuid != nullptr) {
         // Hash for vuid text
         location = XXH32(text_vuid, strlen(text_vuid), 8);
+        if ((debug_data->duplicate_message_limit > 0) && UpdateLogMsgCounts(debug_data, location)) {
+            // Count for this particular message is over the limit, ignore it
+            return false;
+        }
     }
 
     VkDebugUtilsMessengerCallbackDataEXT callback_data;
@@ -419,7 +441,7 @@ static inline bool debug_log_msg(const debug_report_data *debug_data, VkFlags ms
         oss << "[ " << text_vuid << " ] ";
     }
     uint32_t index = 0;
-    for (auto src_object : object_name_info) {
+    for (const auto &src_object : object_name_info) {
         if (0 != src_object.objectHandle) {
             oss << "Object " << index++ << ": handle = 0x" << std::hex << src_object.objectHandle;
             if (src_object.pObjectName) {
@@ -438,11 +460,11 @@ static inline bool debug_log_msg(const debug_report_data *debug_data, VkFlags ms
     const auto callback_list = &debug_data->debug_callback_list;
     // We only output to default callbacks if there are no non-default callbacks
     bool use_default_callbacks = true;
-    for (auto current_callback : *callback_list) {
+    for (const auto &current_callback : *callback_list) {
         use_default_callbacks &= current_callback.IsDefault();
     }
 
-    for (auto current_callback : *callback_list) {
+    for (const auto &current_callback : *callback_list) {
         // Skip callback if it's a default callback and there are non-default callbacks present
         if (current_callback.IsDefault() && !use_default_callbacks) continue;
 
@@ -571,14 +593,14 @@ static inline VkResult layer_create_report_callback(debug_report_data *debug_dat
 static inline void ActivateInstanceDebugCallbacks(debug_report_data *debug_data) {
     auto current = debug_data->instance_pnext_chain;
     for (;;) {
-        auto create_info = lvl_find_in_chain<VkDebugUtilsMessengerCreateInfoEXT>(current);
+        auto create_info = LvlFindInChain<VkDebugUtilsMessengerCreateInfoEXT>(current);
         if (!create_info) break;
         current = create_info->pNext;
         VkDebugUtilsMessengerEXT utils_callback{};
         layer_create_callback((DEBUG_CALLBACK_UTILS | DEBUG_CALLBACK_INSTANCE), debug_data, create_info, nullptr, &utils_callback);
     }
     for (;;) {
-        auto create_info = lvl_find_in_chain<VkDebugReportCallbackCreateInfoEXT>(current);
+        auto create_info = LvlFindInChain<VkDebugReportCallbackCreateInfoEXT>(current);
         if (!create_info) break;
         current = create_info->pNext;
         VkDebugReportCallbackEXT report_callback{};
@@ -587,12 +609,12 @@ static inline void ActivateInstanceDebugCallbacks(debug_report_data *debug_data)
 }
 
 static inline void DeactivateInstanceDebugCallbacks(debug_report_data *debug_data) {
-    if (!lvl_find_in_chain<VkDebugUtilsMessengerCreateInfoEXT>(debug_data->instance_pnext_chain) &&
-        !lvl_find_in_chain<VkDebugReportCallbackCreateInfoEXT>(debug_data->instance_pnext_chain))
+    if (!LvlFindInChain<VkDebugUtilsMessengerCreateInfoEXT>(debug_data->instance_pnext_chain) &&
+        !LvlFindInChain<VkDebugReportCallbackCreateInfoEXT>(debug_data->instance_pnext_chain))
         return;
     std::vector<VkDebugUtilsMessengerEXT> instance_utils_callback_handles{};
     std::vector<VkDebugReportCallbackEXT> instance_report_callback_handles{};
-    for (auto item : debug_data->debug_callback_list) {
+    for (const auto &item : debug_data->debug_callback_list) {
         if (item.IsInstance()) {
             if (item.IsUtils()) {
                 instance_utils_callback_handles.push_back(item.debug_utils_callback_object);
@@ -601,10 +623,10 @@ static inline void DeactivateInstanceDebugCallbacks(debug_report_data *debug_dat
             }
         }
     }
-    for (auto item : instance_utils_callback_handles) {
+    for (const auto &item : instance_utils_callback_handles) {
         layer_destroy_callback(debug_data, item, nullptr);
     }
-    for (auto item : instance_report_callback_handles) {
+    for (const auto &item : instance_report_callback_handles) {
         layer_destroy_callback(debug_data, item, nullptr);
     }
 }
@@ -628,22 +650,62 @@ static inline bool LogMsgLocked(const debug_report_data *debug_data, VkFlags msg
                                 const std::string &vuid_text, char *err_msg) {
     std::string str_plus_spec_text(err_msg ? err_msg : "Allocation failure");
 
+    // If message is in filter list, bail out very early
+    size_t message_id = XXH32(vuid_text.c_str(), strlen(vuid_text.c_str()), 8);
+    if (std::find(debug_data->filter_message_ids.begin(), debug_data->filter_message_ids.end(),
+                  static_cast<uint32_t>(message_id)) != debug_data->filter_message_ids.end())
+        return false;
+
     // Append the spec error text to the error message, unless it's an UNASSIGNED or UNDEFINED vuid
-    if ((vuid_text.find("UNASSIGNED-") == std::string::npos) && (vuid_text.find(kVUIDUndefined) == std::string::npos)) {
+    if ((vuid_text.find("UNASSIGNED-") == std::string::npos) && (vuid_text.find(kVUIDUndefined) == std::string::npos) &&
+        (vuid_text.rfind("SYNC-", 0) == std::string::npos)) {
         // Linear search makes no assumptions about the layout of the string table. This is not fast, but it does not need to be at
         // this point in the error reporting path
         uint32_t num_vuids = sizeof(vuid_spec_text) / sizeof(vuid_spec_text_pair);
         const char *spec_text = nullptr;
+        std::string spec_type;
         for (uint32_t i = 0; i < num_vuids; i++) {
             if (0 == strcmp(vuid_text.c_str(), vuid_spec_text[i].vuid)) {
                 spec_text = vuid_spec_text[i].spec_text;
+                spec_type = vuid_spec_text[i].url_id;
                 break;
             }
         }
 
+        // Construct and append the specification text and link to the appropriate version of the spec
         if (nullptr != spec_text) {
-            str_plus_spec_text += " The Vulkan spec states: ";
-            str_plus_spec_text += spec_text;
+            std::string spec_link = "https://www.khronos.org/registry/vulkan/specs/_MAGIC_KHRONOS_SPEC_TYPE_/html/vkspec.html";
+#ifdef ANNOTATED_SPEC_LINK
+            spec_link = ANNOTATED_SPEC_LINK;
+#endif
+            static std::string kAtToken = "_MAGIC_ANNOTATED_SPEC_TYPE_";
+            static std::string kKtToken = "_MAGIC_KHRONOS_SPEC_TYPE_";
+            static std::string kVeToken = "_MAGIC_VERSION_ID_";
+            auto Replace = [](std::string &dest_string, const std::string &to_replace, const std::string &replace_with) {
+                if (dest_string.find(to_replace) != std::string::npos) {
+                    dest_string.replace(dest_string.find(to_replace), to_replace.size(), replace_with);
+                }
+            };
+
+            str_plus_spec_text.append(" The Vulkan spec states: ");
+            str_plus_spec_text.append(spec_text);
+            if (0 == spec_type.compare("default")) {
+                str_plus_spec_text.append(" (https://github.com/KhronosGroup/Vulkan-Docs/search?q=)");
+            } else {
+                str_plus_spec_text.append(" (");
+                str_plus_spec_text.append(spec_link);
+                std::string major_version = std::to_string(VK_VERSION_MAJOR(VK_HEADER_VERSION_COMPLETE));
+                std::string minor_version = std::to_string(VK_VERSION_MINOR(VK_HEADER_VERSION_COMPLETE));
+                std::string patch_version = std::to_string(VK_VERSION_PATCH(VK_HEADER_VERSION_COMPLETE));
+                std::string header_version = major_version + "." + minor_version + "." + patch_version;
+                std::string annotated_spec_type = major_version + "." + minor_version + "-extensions";
+                Replace(str_plus_spec_text, kKtToken, spec_type);
+                Replace(str_plus_spec_text, kAtToken, annotated_spec_type);
+                Replace(str_plus_spec_text, kVeToken, header_version);
+                str_plus_spec_text.append("#");  // CMake hates hashes
+            }
+            str_plus_spec_text.append(vuid_text);
+            str_plus_spec_text.append(")");
         }
     }
 
